"""Provide Jinja2 Helpers."""
from jinja2 import BaseLoader, ChoiceLoader, Environment, TemplateNotFound
from pkg_resources import resource_string
from starlette.templating import Jinja2Templates as _Jinja2Templates


class PkgResourcesTemplateLoader(BaseLoader):
    """Load jinja templates via pkg_resources."""

    @staticmethod
    def get_source(environment: Environment, template: str):
        """Load the template via package resources."""
        source = None
        if ':' in template:
            pkg, resource = template.split(':')
            try:
                source = resource_string(pkg, resource).decode()
            except FileNotFoundError:
                pass
        if source is None:
            raise TemplateNotFound(template)
        return source, None, lambda: True


class Jinja2Templates(_Jinja2Templates):  # noqa: D101
    __doc__ = _Jinja2Templates.__doc__

    def get_env(self, directory: str) -> Environment:
        """Get the environment."""
        env = super().get_env(directory)
        # We want to setup the choice loader here
        env.loader = ChoiceLoader([PkgResourcesTemplateLoader(), env.loader])
        return env



"""Base validator for token based authentication."""
from authlib.jose import errors as jwt_errors
from fastapi.security import OAuth2AuthorizationCodeBearer
from fastapi.security.utils import get_authorization_scheme_param
from pydantic import BaseModel
from starlette.middleware.authentication import AuthenticationError
from starlette.requests import Request

from fastapi_aad_auth._base.state import AuthenticationState, User
from fastapi_aad_auth._base.validators.base import Validator


class InitOAuth(BaseModel):
    """OAuth information for openapi docs."""
    clientId: str
    scopes: str
    usePkceWithAuthorizationCodeGrant: bool


class TokenValidator(Validator, OAuth2AuthorizationCodeBearer):  # type: ignore
    """Validator for token based authentication."""

    def __init__(
        self,
        client_id: str,
        authorizationUrl: str,
        tokenUrl: str,
        api_audience: str = None,
        scheme_name: str = None,
        scopes: dict = None,
        auto_error: bool = False,
        enabled: bool = True,
        use_pkce: bool = True,
        user_klass: type = User
    ):
        """Initialise validator for token based authentication."""
        super().__init__(authorizationUrl=authorizationUrl, tokenUrl=tokenUrl, refreshUrl=api_audience, scheme_name=scheme_name, scopes=scopes, auto_error=auto_error)
        self.client_id = client_id
        self.enabled = enabled
        if api_audience is None:
            api_audience = f"api://{client_id}"
        self.api_audience = api_audience
        self._use_pkce = use_pkce
        self._user_klass = user_klass

    def check(self, request: Request):
        """Check the authentication from the request."""
        state = AuthenticationState.as_unauthenticated(None, None)
        try:
            token = self.get_token(request)
            if token is not None:
                claims = self.validate_token(token)
                user = self._get_user_from_claims(claims)
                state = AuthenticationState.authenticate_as(user, None, None)
        except Exception:
            self.logger.exception('Error authenticating via token')
        return state

    def get_token(self, request: Request):
        """Get the token from the request."""
        authorization = request.headers.get("Authorization")
        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            if self.auto_error:
                raise self.not_authenticated
            else:
                return None  # pragma: nocover
        return param

    @property
    def init_oauth(self):
        """Get the openapi docs config."""
        return InitOAuth(clientId=self.client_id, scopes=f'{self.api_audience}/openid', usePkceWithAuthorizationCodeGrant=self._use_pkce).dict()

    def _validate_claims(self, claims, options=None):
        if options is None:
            options = self._claims_options
        claims.options = options
        try:
            claims.validate()
        except jwt_errors.ExpiredTokenError as e:
            self.logger.error(f'Expired token:\n\t{self._compare_claims(claims)}')
            raise AuthenticationError(f"Token is expired {e.args}")
        except jwt_errors.InvalidClaimError as e:
            self.logger.error(f'Invalid claims:\n\t{self._compare_claims(claims)}')
            raise AuthenticationError(f"Invalid claims {e.args}")
        except jwt_errors.MissingClaimError as e:
            self.logger.error(f'Missing claims:\n\t{self._compare_claims(claims)}')
            raise AuthenticationError(f"Missing claims {e.args}")
        except Exception as e:
            self.logger.exception('Unable to parse error')
            raise AuthenticationError(f"Unable to parse authentication token {e.args}")
        return claims

    @property
    def _claims_options(self):
        options = {"sub": {"essential": True},
                   "aud": {"essential": True, "values": [self.api_audience]},
                   "exp": {"essential": True},
                   "nbf": {"essential": True},
                   "iat": {"essential": True}}
        return options

    def _decode_token(self, token):
        raise NotImplementedError('Implement in base class')

    def validate_token(self, token, options=None):
        """Validate provided token."""
        claims = self._decode_token(token)
        return self._validate_claims(claims, options)

    @staticmethod
    def _compare_claims(claims):
        return '\n\t'.join([f'{key}: {value} - {claims.options.get(key, None)}' for key, value in claims.items()])

    def _get_user_from_claims(self, claims):
        raise NotImplementedError('Implement in sub class')



"""fastapi_aad_auth configuration options."""
from typing import Dict, List, Optional, Union
import uuid

from pkg_resources import resource_filename
from pydantic import BaseSettings as _BaseSettings, DirectoryPath, Field, FilePath, SecretStr, validator

from fastapi_aad_auth.providers.aad import AADConfig
from fastapi_aad_auth.utilities import bool_from_env, DeprecatableFieldsMixin, DeprecatedField, expand_doc, klass_from_str


class BaseSettings(DeprecatableFieldsMixin, _BaseSettings):
    """Allow deprecations in the BaseSettings object."""


_DEPRECATION_VERSION = '0.2.0'


@expand_doc
class RoutingConfig(BaseSettings):
    """Configuration for authentication related routing.

    Includes ``logout_path``, ``login_path``, and ``login_redirect_path`` (defaults should
    be fine for most use-cases).

    There are also the ``landing_path`` for the login page, as well as the ``home_path`` for the home
    page (defaults to the application root), and the ``post_logout_path`` for any specific routing
    once a logout has completed.
    """

    login_path: str = DeprecatedField('/login/oauth', description="Path for initiating the AAD oauth call", env='FASTAPI_AUTH_LOGIN_ROUTE', deprecated_in=_DEPRECATION_VERSION, replaced_by='Routing.oauth_base_route', additional_info=' - To access the new behaviour, set this value to None or an empty string')
    login_redirect_path: str = DeprecatedField('/login/oauth/redirect', description="Path for handling the AAD redirect call", env='FASTAPI_AUTH_LOGIN_REDIRECT_ROUTE', deprecated_in=_DEPRECATION_VERSION, replaced_by='Routing.oauth_base_route', additional_info=' - To access the new behaviour, set this value to None or an empty string')
    oauth_base_route: str = Field('/oauth', description="Base Path for initiating the oauth calls", env='FASTAPI_OAUTH_BASE_ROUTE')
    logout_path: str = Field('/logout', description="Path for processing a logout request", env='FASTAPI_AUTH_LOGOUT_ROUTE')
    landing_path: str = Field('/login', description="Path for the login UI page", env='FASTAPI_AUTH_LOGIN_UI_ROUTE')
    user_path: Optional[str] = Field('/me', description="Path for getting the user view", env='FASTAPI_AUTH_USER_ROUTE')
    home_path: str = Field('/', description="Path for the application home page (default redirect if none provided)",
                           env='APP_HOME_ROUTE')
    post_logout_path: str = Field(None, description="Path for the redirect post logout - defaults to the landing path if not provided",
                                  env='FASTAPI_AUTH_POST_LOGOUT_ROUTE')

    class Config:  # noqa D106
        env_file = '.env'

    @validator('post_logout_path', always=True, pre=True)
    def _validate_post_logout_path(cls, value, values):
        if value is None:
            value = values.get('landing_path')
        return value

    @property
    def no_redirect_routes(self):
        """Routes that we don't want to redirect to."""
        return [self.login_path, self.login_redirect_path, f'{self.oauth_base_route}/*']


@expand_doc
class LoginUIConfig(BaseSettings):
    """Configuration for the application Login UI.

    Includes the application name, template file, error template file,
    static directory, path to mount the login static information to, and the context.
    """
    app_name: str = Field(None, description="Application name to show on the Login UI page", env='APP_NAME')
    template_file: FilePath = Field(resource_filename('fastapi_aad_auth.ui', 'login.html'),
                                    description="The jinja2 template to use for the login screen",
                                    env='FASTAPI_AUTH_LOGIN_TEMPLATE_FILE')
    error_template_file: FilePath = Field(resource_filename('fastapi_aad_auth.ui', 'error.html'),
                                          description="The jinja2 template to use for error information",
                                          env='FASTAPI_AUTH_LOGIN_ERROR_TEMPLATE_FILE')
    user_template_file: FilePath = Field(resource_filename('fastapi_aad_auth.ui', 'user.html'),
                                         description="The jinja2 template to use for the user view",
                                         env='FASTAPI_AUTH_USER_TEMPLATE_FILE')
    static_directory: DirectoryPath = Field(resource_filename('fastapi_aad_auth.ui', 'static'),
                                            description="Static path for the UI components",
                                            env='FASTAPI_AUTH_LOGIN_STATIC_DIR')
    static_path: str = Field('/static-login',
                             description="Path to mount the login static dir in",
                             env='FASTAPI_AUTH_LOGIN_STATIC_PATH')
    context: Optional[Dict[str, str]] = Field(None, description="Any additional context variables required for the template")
    ui_klass: type = Field('fastapi_aad_auth.ui:UI',
                           description="UI class to use to handle creating and returning the routes for the login, error and user screens, this will be treated as an import path "
                           "if provided as a string, with the last part the class to load", env='FASTAPI_AUTH_UI_KLASS')

    class Config:  # noqa D106
        env_file = '.env'

    _validate_klass = validator('ui_klass', pre=True, always=True, allow_reuse=True)(klass_from_str)


@expand_doc
class AuthSessionConfig(BaseSettings):
    """Authentication Session configuration.

    Contains secret and salt information (should be set as environment
    variables in a multi-worker/multi-processing environment to enable
    authentication across workers)
    """
    secret: SecretStr = Field(str(uuid.uuid4()), description="Secret used for encoding authentication information",
                              env='SESSION_AUTH_SECRET')
    salt: SecretStr = Field(str(uuid.uuid4()), description="Salt used for encoding authentication information",
                            env='SESSION_AUTH_SALT')

    class Config:  # noqa D106
        env_file = '.env'


@expand_doc
class SessionConfig(BaseSettings):
    """Configuration for session middleware.

    Contains the session secret (should be set as environment variables in
    a multi-worker/multi-processing environment to enable authentication
    across workers)

    Provides configuration for the fastapi session middleware
    """
    secret_key: SecretStr = Field(str(uuid.uuid4()), description="Secret used for the session middleware",
                                  env='SESSION_SECRET')
    session_cookie: str = Field('session', description="Cookie name for the session information",
                                env='SESSION_COOKIE')
    same_site: str = Field('lax', description="Cookie validation mode for the session", env='SESSION_SAME_SITE')
    https_only: bool = Field(False, description="Allow the sessions only with https connections", env='SESSION_HTTPS_ONLY')
    max_age: int = Field(24*60*60, description="Maximum age for a session", env='SESSION_MAX_AGE')

    class Config:  # noqa D106
        env_file = '.env'

    _validate_https_only = validator('https_only', allow_reuse=True)(bool_from_env)


@expand_doc
class Config(BaseSettings):
    """The overall configuration for the AAD authentication.

    Provides the overall configuration and parameters.
    """

    enabled: bool = Field(True, description="Enable authentication", env='FASTAPI_AUTH_ENABLED')
    providers: List[Union[AADConfig]] = Field(None, description="The provider configurations to use")
    aad: Optional[AADConfig] = DeprecatedField(None, description='AAD Configuration information', deprecated_in='0.2.0', replaced_by='Config.providers')
    auth_session: AuthSessionConfig = Field(None, description="The configuration for encoding the authentication information in the session")
    routing: RoutingConfig = Field(None, description="Configuration for routing")
    session: SessionConfig = Field(None, description="Configuration for the session middleware")
    login_ui: LoginUIConfig = Field(None, description="Login UI Configuration")
    user_klass: type = Field('fastapi_aad_auth._base.state:User',
                             description="User class to use within the AADOAuthBackend, this will be treated as an import path "
                             "if provided as a string, with the last part the class to load", env='FASTAPI_AUTH_USER_KLASS')

    class Config:  # noqa D106
        env_file = '.env'

    @validator('providers', always=True, pre=True)
    def _validate_providers(cls, value, values):
        enabled = values.get('enabled', cls.__fields__['enabled'].default)
        if value is None:
            value = []
            if enabled:
                value.append(AADConfig(_env_file=cls.Config.env_file))
        return value

    @validator('aad', always=True, pre=True)
    def _validate_aad(cls, value, values):
        enabled = values.get('enabled', cls.__fields__['enabled'].default)
        if value is None and enabled:
            providers = values.get('providers', [AADConfig(_env_file=cls.Config.env_file)])
            value = [u for u in providers if isinstance(u, AADConfig)][0]
        return value

    @validator('auth_session', always=True, pre=True)
    def _validate_auth_session(cls, value):
        if value is None:
            value = AuthSessionConfig(_env_file=cls.Config.env_file)
        return value

    @validator('routing', always=True, pre=True)
    def _validate_routing(cls, value):
        if value is None:
            value = RoutingConfig(_env_file=cls.Config.env_file)
        return value

    @validator('session', always=True, pre=True)
    def _validate_session(cls, value):
        if value is None:
            value = SessionConfig(_env_file=cls.Config.env_file)
        return value

    @validator('login_ui', always=True, pre=True)
    def _validate_login_ui(cls, value):
        if value is None:
            value = LoginUIConfig(_env_file=cls.Config.env_file)
        return value

    _validate_klass = validator('user_klass', pre=True, always=True, allow_reuse=True)(klass_from_str)
    _validate_enabled = validator('enabled', allow_reuse=True)(bool_from_env)



"""Authentication State."""
from fastapi_aad_auth._base.state import AuthenticationState as _AuthenticationState, User as _User
from fastapi_aad_auth.utilities import deprecate, deprecate_module

_DEPRECATED_VERSION = '0.2.0'


deprecate_module(locals(), _DEPRECATED_VERSION, replaced_by=_AuthenticationState.__module__)


@deprecate(_DEPRECATED_VERSION, replaced_by=f'{_AuthenticationState.__module__}:{_AuthenticationState.__name__}')
class AuthenticationState(_AuthenticationState):  # noqa: D101
    __doc__ = _AuthenticationState.__doc__


@deprecate(_DEPRECATED_VERSION, replaced_by=f'{_User.__module__}:{_User.__name__}')
class User(_User):  # noqa: D101
    __doc__ = _User.__doc__



"""Authentication State Handler."""
from enum import Enum
import json
from typing import List, Optional
import uuid

from itsdangerous import URLSafeSerializer
from itsdangerous.exc import BadSignature
from pydantic import BaseModel, Field, root_validator, validator
from starlette.authentication import AuthCredentials, SimpleUser, UnauthenticatedUser

from fastapi_aad_auth.errors import AuthenticationError
from fastapi_aad_auth.mixins import LoggingMixin


SESSION_STORE_KEY = 'auth'


class AuthenticationOptions(Enum):
    """Authentication Options."""
    unauthenticated = 0
    not_allowed = -1
    authenticated = 1


class User(BaseModel):
    """User Model."""
    name: str = Field(..., description='Full name')
    email: str = Field(..., description='User email')
    username: str = Field(..., description='Username')
    roles: Optional[List[str]] = Field(None, description='Any roles provided')
    groups: Optional[List[str]] = Field(None, description='Any groups provided')
    scopes: Optional[List[str]] = Field(None, description='Token scopes provided')

    @property
    def permissions(self):
        """User Permissions."""
        permissions = []
        if self.scopes:
            for scope in self.scopes:
                if not scope.startswith('.'):
                    permissions.append(scope)
        return permissions[:]

    @validator('scopes', always=True, pre=True)
    def _validate_scopes(cls, value):
        if isinstance(value, str):
            value = value.split(' ')
        return value


class AuthenticationState(LoggingMixin, BaseModel):
    """Authentication State."""
    session_state: str = str(uuid.uuid4())
    state: AuthenticationOptions = AuthenticationOptions.unauthenticated
    user: Optional[User] = None
    _logger = None

    class Config:  # noqa: D106
        underscore_attrs_are_private = True

    @root_validator(pre=True)
    def _validate_user(cls, values):
        if values.get('user', None) is None:
            values['state'] = AuthenticationOptions.unauthenticated
        return values

    def check_session_state(self, session_state):
        """Check state against session state."""
        if session_state != self.session_state:
            raise AuthenticationError("Session states do not match")
        return True

    def store(self, serializer):
        """Store in serializer."""
        return serializer.dumps(self.json())

    @classmethod
    def load(cls, serializer: URLSafeSerializer, encoded_state: Optional[str] = None):
        """Load from encoded state.

        Args:
            serializer: Serializer object containing the en/decoding secrets
        Keyword Args:
            encoded_state: The encoded state to be decoded
        """
        if encoded_state:
            try:
                state = json.loads(serializer.loads(encoded_state))
                loaded_state = cls(**state)
            except BadSignature:
                loaded_state = cls()
        else:
            loaded_state = cls()
        return loaded_state

    @classmethod
    def logout(cls, serializer: URLSafeSerializer, session):
        """Clear the sessions state."""
        state = cls.load_from_session(serializer, session)
        state.user = None
        state.state = AuthenticationOptions.unauthenticated
        session[SESSION_STORE_KEY] = state.store(serializer)

    @classmethod
    def load_from_session(cls, serializer: URLSafeSerializer, session):
        """Load from a session."""
        return cls.load(serializer, session.get(SESSION_STORE_KEY, None))

    def save_to_session(self, serializer: URLSafeSerializer, session):
        """Save to a session."""
        session[SESSION_STORE_KEY] = self.store(serializer)
        return session

    def is_authenticated(self):
        """Check if the state is authenticated."""
        return self.user is not None and self.state == AuthenticationOptions.authenticated

    @property
    def authenticated_user(self):
        """Get the authenticated user."""
        if self.is_authenticated() and self.user:
            if isinstance(self.user, User):
                return SimpleUser(self.user.email)
        return UnauthenticatedUser()

    @property
    def credentials(self):
        """Get the credentials object."""
        if self.user and self.is_authenticated():
            return AuthCredentials(['authenticated'] + self.user.permissions)
        else:
            return AuthCredentials()

    @classmethod
    def authenticate_as(cls, user, serializer, session):
        """Store the authenticated user."""
        state = cls(user=user, state=AuthenticationOptions.authenticated)
        if serializer is not None and session is not None:
            state.save_to_session(serializer, session)
        return state

    @classmethod
    def as_unauthenticated(cls, serializer, session):
        """Store as an un-authenticated user."""
        return cls.authenticate_as(None, serializer, session)



"""Authenticator Class."""
from functools import wraps
from pathlib import Path
from typing import Any, Dict, List, Optional

from fastapi import FastAPI
from starlette.authentication import requires
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.requests import Request
from starlette.responses import RedirectResponse, Response
from starlette.routing import request_response, Route

from fastapi_aad_auth._base.backend import BaseOAuthBackend
from fastapi_aad_auth._base.validators import SessionValidator
from fastapi_aad_auth.config import Config
from fastapi_aad_auth.errors import AuthenticationError, AuthorisationError, base_error_handler, ConfigurationError, json_error_handler, redirect_error_handler
from fastapi_aad_auth.mixins import LoggingMixin
from fastapi_aad_auth.ui.jinja import Jinja2Templates
from fastapi_aad_auth.utilities import deprecate, is_interactive


_BASE_ROUTES = ['openapi', 'swagger_ui_html', 'swagger_ui_redirect', 'redoc_html']


class Authenticator(LoggingMixin):
    """Authenticator class.

    Creates the key components based on the provided configurations
    """

    def __init__(self, config: Config = None, add_to_base_routes: bool = True, base_context: Optional[Dict[str, Any]] = None, user_klass: Optional[type] = None):
        """Initialise the Authenticator based on the provided configuration.

        Keyword Args:
            config (fastapi_aad_auth.config.Config): Authentication configuration (includes ui and routing, as well as AAD Application and Tenant IDs)
            add_to_base_routes (bool): Add the authentication to the router
            base_context (Dict[str, Any]): a base context to provide
            user_klass (type): The user class to use as part of the auth state
        """
        super().__init__()
        if config is None:
            config = Config()
        if user_klass is not None:
            config.user_klass = user_klass
        self.config = config
        if base_context is None:
            base_context = {}
        if self.config.login_ui.context:
            context = self.config.login_ui.context.copy()
            context.update(base_context)
            base_context = context
        if 'app_name' not in base_context:
            base_context['app_name'] = self.config.login_ui.app_name
        if 'static_path' not in base_context:
            base_context['static_path'] = self.config.login_ui.static_path
        self._base_context = base_context
        self._add_to_base_routes = add_to_base_routes
        self._session_validator = self._init_session_validator()
        self._providers = self._init_providers()
        self.auth_backend = self._init_auth_backend()
        self._ui = None
        self._ui_routes = self._init_ui()
        self._auth_routes = self._init_auth_routes()

    def _init_session_validator(self):
        auth_serializer = SessionValidator.get_session_serializer(self.config.auth_session.secret.get_secret_value(),
                                                                  self.config.auth_session.salt.get_secret_value())
        return SessionValidator(auth_serializer, ignore_redirect_routes=self.config.routing.no_redirect_routes)
        # Lets setup the oauth backend

    def _init_providers(self):
        return [u._provider_klass.from_config(session_validator=self._session_validator, config=self.config, provider_config=u) for u in self.config.providers]

    def _init_auth_backend(self):
        validators = [self._session_validator]
        for provider in self._providers:
            validators += provider.validators
        return BaseOAuthBackend(validators, enabled=self.config.enabled)

    def _init_ui(self):
        self._ui = self.config.login_ui.ui_klass(self.config, self, self._base_context)
        return self._ui.routes

    def _init_auth_routes(self):

        async def logout(request: Request):
            self.logger.debug(f'Logging out - request url {request.url}')
            if self.config.enabled:
                self.logger.debug(f'Auth {request.auth}')
                for provider in self._providers:
                    provider.logout(request)
                self._session_validator.logout(request)
            return RedirectResponse(self.config.routing.post_logout_path)
        routes = [Route(self.config.routing.logout_path, endpoint=logout, methods=['GET'], name='logout')]
        for provider in self._providers:
            routes += provider.get_routes(noauth_redirect=self.config.routing.home_path)
        # We have a deprecated behaviour here
        return routes

    def _set_error_handlers(self, app):
        error_template_path = Path(self.config.login_ui.error_template_file)
        error_templates = Jinja2Templates(directory=str(error_template_path.parent))
        if self.config.login_ui.app_name:
            self._base_context['appname'] = self.config.login_ui.app_name
        else:
            self._base_context['appname'] = app.title
        self._base_context['static_path'] = self.config.login_ui.static_path

        @app.exception_handler(ConfigurationError)
        async def configuration_error_handler(request: Request, exc: ConfigurationError) -> Response:
            error_message = "Oops! It seems like the application has not been configured correctly, please contact an admin"
            error_type = 'Authentication Configuration Error'
            status_code = 500
            return base_error_handler(request, exc, error_type, error_message, error_templates, error_template_path, context=self._base_context.copy(), status_code=status_code)

        @app.exception_handler(AuthorisationError)
        async def authorisation_error_handler(request: Request, exc: AuthorisationError) -> Response:
            error_message = "Oops! It seems like you cannot access this information. If this is an error, please contact an admin"
            error_type = 'Authorisation Error'
            status_code = 403
            return base_error_handler(request, exc, error_type, error_message, error_templates, error_template_path, context=self._base_context.copy(), status_code=status_code)

        @app.exception_handler(AuthenticationError)
        async def authentication_error_handler(request: Request, exc: AuthenticationError) -> Response:
            return self._authentication_error_handler(request, exc)

    def _authentication_error_handler(self, request: Request, exc: AuthenticationError) -> Response:
        error_message = "Oops! It seems like you are not correctly authenticated"
        status_code = 401
        self.logger.exception(f'Error {exc} for request {request}')
        if is_interactive(request):
            self._session_validator.set_post_auth_redirect(request, request.url.path)
            kwargs = {}
            if self._session_validator.is_valid_redirect(request.url.path):
                kwargs['redirect'] = request.url.path
            return redirect_error_handler(self.config.routing.landing_path, exc, **kwargs)
        else:
            return json_error_handler(error_message, status_code=status_code)

    def auth_required(self, scopes: str = 'authenticated', redirect: str = 'login'):
        """Decorator to require specific scopes (and redirect to the login ui) for an endpoint.

        This can be used for toggling authentication (e.g. between an internal/external server)
        as well as handling the redirection based on the session information

        Keyword Args:
            scopes: scopes for the fastapi requires decorator
            redirect: name of the redirection url
        """

        def wrapper(endpoint):
            if self.config.enabled:

                @wraps(endpoint)
                async def require_endpoint(request: Request, *args, **kwargs):
                    self._session_validator.set_post_auth_redirect(request, request.url.path)

                    @requires(scopes, redirect=redirect)
                    async def req_wrapper(request: Request, *args, **kwargs):
                        return await endpoint(request, *args, **kwargs)

                    return await req_wrapper(request, *args, **kwargs)

                return require_endpoint
            else:
                return endpoint

        return wrapper

    def app_routes_add_auth(self, app: FastAPI, route_list: List[str], invert: bool = False):
        """Add authentication to specified routes in application router.

        Used for default routes (e.g. api/docs and api/redocs, openapi.json etc)

        Args:
            app: fastapi application
            route_list: list of routes to add authentication to (e.g. api docs, redocs etc)

        Keyword Args:
            invert: Switch between using the route list as a block list or an allow list

        """
        if self.config.enabled:
            routes = app.router.routes
            for i, route in enumerate(routes):
                # Can use allow list or block list (i.e. invert = True sets all except the route list to have auth
                if (route.name in route_list and not invert) or (route.name not in route_list and invert):  # type: ignore
                    route.endpoint = self.auth_required()(route.endpoint)  # type: ignore
                    route.app = request_response(route.endpoint)  # type: ignore
                app.router.routes[i] = route
        return app

    def configure_app(self, app: FastAPI, add_error_handlers=True):
        """Configure the fastapi application to use these authentication handlers.

        Adds authentication middleware, error handler and adds authentication
        to the default routes as well as adding the authentication specific routes

        Args:
            app: fastapi application

        Keyword Args:
            add_error_handlers (bool) : add the error handlers to the app (default is true, but can be set to False to configure specific handling)
        """
        def on_auth_error(request: Request, exc: AuthenticationError):
            return self._authentication_error_handler(request, exc)

        app.add_middleware(AuthenticationMiddleware, backend=self.  auth_backend, on_error=on_auth_error)
        if add_error_handlers:
            self._set_error_handlers(app)
        # Check if session middleware is there
        if not any([SessionMiddleware in u.cls.__mro__ for u in app.user_middleware]):
            app.add_middleware(SessionMiddleware, **self.config.session.dict())
        if self._add_to_base_routes:
            self.app_routes_add_auth(app, _BASE_ROUTES)
        app.routes.extend(self._ui_routes)
        app.routes.extend(self._auth_routes)
        # TODO: select a specific provider to use here
        if self.config.enabled:
            app.swagger_ui_init_oauth = self._providers[0].validators[0].init_oauth


_DEPRECATED_VERSION = '0.2.0'


@deprecate(_DEPRECATED_VERSION, replaced_by=f'{Authenticator.__module__}:{Authenticator.__name__}')
class AADAuth(Authenticator):   # noqa: D101
    __doc__ = Authenticator.__doc__

    @property  # type: ignore
    @deprecate(_DEPRECATED_VERSION, replaced_by=f'{Authenticator.__module__}:{Authenticator.__name__}.auth_backend.requires_auth')
    def api_auth_scheme(self):
        """Get the API Authentication Schema."""
        return self.auth_backend.requires_auth()



"""UI Components and templates.

Includes:

    * ``static``: static content (css/js/fonts) for the ui components
    * ``error.html``: Template for error messages
    * ``login.html``: Login page for UI login
    * ``user.html``: View the user's information and get an access token
"""
from pathlib import Path
from typing import Any, Dict, List, Optional

from fastapi import Depends
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse
from starlette.routing import Mount, Route
from starlette.staticfiles import StaticFiles


from fastapi_aad_auth import auth, config  # noqa: F401
from fastapi_aad_auth._base.state import AuthenticationState
from fastapi_aad_auth.mixins import LoggingMixin
from fastapi_aad_auth.ui.jinja import Jinja2Templates
from fastapi_aad_auth.utilities import urls


class UI(LoggingMixin):
    """Provides Login endpoint methods, which are then wrapped in a factory method."""
    def __init__(self, config: 'config.Config', authenticator: 'auth.Authenticator', base_context: Dict[str, Any] = None):
        """Initialise the UI based on the provided configuration.

        Keyword Args:
            config (fastapi_aad_auth.config.Config): Authentication configuration (includes ui and routing, as well as AAD Application and Tenant IDs)
            authenticator (fastapi_aad_auth.auth.Authenticator): The authenticator object
            base_context (Dict[str, Any]): Add the authentication to the router
        """
        super().__init__()
        self.config = config
        if base_context is None:
            base_context = dict()
        self._base_context = base_context
        self._authenticator = authenticator

        self.login_template_path = Path(self.config.login_ui.template_file)
        self.user_template_path = Path(self.config.login_ui.user_template_file)
        self.login_templates = Jinja2Templates(directory=str(self.login_template_path.parent))
        self.user_templates = Jinja2Templates(directory=str(self.user_template_path.parent))

    def _login(self, request: Request, *args, **kwargs):
        """Provide the Login UI."""
        if not self.config.enabled or self._authenticator.auth_backend.is_authenticated(request):
            return RedirectResponse(self.config.routing.home_path)
        context = self._base_context.copy()
        context.update(kwargs)  # type: ignore
        if not self.config.enabled or request.user.is_authenticated:
            # This is authenticated so go straight to the homepage
            return RedirectResponse(self.config.routing.home_path)
        context['request'] = request  # type: ignore
        if 'login' not in context or context['login'] is None:  # type: ignore
            post_redirect = self._authenticator._session_validator.pop_post_auth_redirect(request)
            context['login'] = '<br>'.join([provider.get_login_button(post_redirect) for provider in self._authenticator._providers])  # type: ignore
        return self.login_templates.TemplateResponse(self.login_template_path.name, context)  # type: ignore

    def _get_user(self, request: Request, **kwargs):
        """Provide a UI with information on the user."""
        if not self.config.enabled:
            return RedirectResponse(self.config.routing.home_path)
        context = self._base_context.copy()  # type: ignore
        context.update(kwargs)
        self.logger.debug(f'Getting token for {request.user}')
        context['request'] = request  # type: ignore
        context['token_api_path'] = f'{self.config.routing.user_path}/token'
        if self.config.enabled:
            self.logger.debug(f'Auth {request.auth}')
            try:
                context['user'] = self._authenticator._session_validator.get_state_from_session(request).user
            except ValueError:
                # If we have one provider, we can force the login, otherwise...
                return self.__force_authenticate(request)
        else:
            self.logger.debug('Auth not enabled')
            context['token_api_path'] = None  # type: ignore
        return self.user_templates.TemplateResponse(self.user_template_path.name, context)

    def _get_token(self, request: Request, auth_state: AuthenticationState, scopes: Optional[List[str]] = None):
        """Return the access token for the user."""
        if not isinstance(auth_state, AuthenticationState):
            user = self.__get_user_from_request(request)
        else:
            user = auth_state.user
        if hasattr(user, 'username'):  # type: ignore
            if scopes is None:
                scopes = request.query_params.get('scopes', None)
            if isinstance(scopes, str):
                scopes = scopes.split(' ')  # type: ignore
            access_token = self.__get_access_token(user, scopes)
            if access_token:
                # We want to get the token for each provider that is authenticated
                return JSONResponse(access_token)   # type: ignore
            else:
                if any([u in request.headers['user-agent'] for u in ['Mozilla', 'Gecko', 'Trident', 'WebKit', 'Presto', 'Edge', 'Blink']]):
                    # If we have one provider, we can force the login, otherwise we need to request which login route
                    return self.__force_authenticate(request)
                else:
                    return JSONResponse('Unable to access token as user has not authenticated via session')
        redirect = '/me/token'
        if scopes:
            self.logger.debug(f'Getting Access Token with scopes {scopes}')
            redirect = urls.with_query_params(redirect, scopes=scopes)
        return RedirectResponse(urls.with_query_params(self.config.routing.landing_path, redirect=redirect))

    @property
    def routes(self):
        """Return the routes for the UI.

        Provides the login UI route, and if the routing config has
        the ``user_path`` set, it also provides the user description
        view (and token endpoint)
        """
        async def login(request: Request, *args, **kwargs):
            return self._login(request)

        routes = [Route(self.config.routing.landing_path, endpoint=login, methods=['GET'], name='login'),
                  Mount(self.config.login_ui.static_path, StaticFiles(directory=str(self.config.login_ui.static_directory)), name='static-login')]

        if self.config.routing.user_path:

            @self._authenticator.auth_required()
            async def get_user(request: Request):
                return self._get_user(request)

            async def get_token(request: Request, auth_state: AuthenticationState = Depends(self._authenticator.auth_backend.requires_auth(allow_session=True)), scopes: Optional[List[str]] = None):
                return self._get_token(request, auth_state, scopes)

            routes += [Route(self.config.routing.user_path, endpoint=get_user, methods=['GET'], name='user'),
                       Route(f'{self.config.routing.user_path}/token', endpoint=get_token, methods=['GET'], name='get-token')]

        return routes

    def __force_authenticate(self, request: Request):
        # lets get the full redirect including any query parameters
        redirect = urls.with_query_params(request.url.path, **request.query_params)
        self.logger.debug(f'Request {request.url}')
        self.logger.info(f'Forcing authentication with redirect = {redirect}')
        if len(self._authenticator._providers) == 1:
            return self._authenticator._providers[0].authenticator.process_login_request(request, force=True, redirect=redirect)
        else:
            return RedirectResponse(urls.with_query_params(self.config.routing.landing_path, redirect=redirect))

    def __get_access_token(self, user, scopes=None):
        access_token = None
        for provider in self._authenticator._providers:
            try:
                access_token = provider.authenticator.get_access_token(user, scopes)
            except ValueError:
                pass
            if access_token is not None:
                break
        return access_token

    def __get_user_from_request(self, request: Request):
        if hasattr(request.user, 'username'):
            user = request.user
        else:
            auth_state = self._authenticator.auth_backend.check(request)
            user = auth_state.user
        return user



"""Method, class, module and Field Deprecation handlers."""
from functools import wraps
import warnings

from pkg_resources import parse_version
from pydantic import Field

from fastapi_aad_auth._version import get_versions


__version__ = get_versions()['version']
del get_versions

BASE_VERSION = parse_version(parse_version(__version__).base_version)  # type: ignore


class APIDeprecationWarning(FutureWarning):
    """Warning when an API component is being deprecated."""


class DeprecatedError(Exception):
    """Warning for developers when a component is planned to be deprecated."""


@wraps(Field)
def DeprecatedField(*args, **kwargs):  # noqa: D103
    deprecated_in = kwargs.get('deprecated_in')
    kwargs['warn_from'] = kwargs.get('warn_from', __version__)
    replaced_by = kwargs.get('replaced_by', None)
    description = kwargs.get('description', '')
    additional_info = kwargs.get('additional_info', '')
    if not description:
        description = ''
        sep = ''
    else:
        sep = ' '
    deprecation_message = _get_deprecation_message('Field', deprecated_in, replaced_by, additional_info)
    description += sep + deprecation_message
    kwargs['description'] = description
    kwargs['deprecated'] = True
    return Field(*args, **kwargs)


class DeprecatableFieldsMixin:
    """Mixin for deprecatable fields."""
    def __new__(cls, *args, **kwargs):
        """Initialise the Field Deprecation Validator."""
        for field_name, field in cls.__fields__.items():
            if field.field_info.extra.get('deprecated', False):
                if field.pre_validators is None:
                    field.pre_validators = []
                field.pre_validators.insert(0, cls._deprecator_validator)
        return super().__new__(cls)

    @staticmethod
    def _deprecator_validator(cls, value, kw, field, *args, **kwargs):
        if field.field_info.extra.get('deprecated', False):
            deprecated_object_description = f'{cls.__module__}:{cls.__name__}.{field.name}'
            env = field.field_info.extra.get('env', None)
            if env:
                deprecated_object_description += f' (env={env})'
            deprecated_in = field.field_info.extra.get('deprecated_in')
            warn_from = field.field_info.extra.get('warn_from', __version__)
            replaced_by = field.field_info.extra.get('replaced_by', None)
            deprecation_message = _get_deprecation_message(deprecated_object_description, deprecated_in, replaced_by)
            _warn(deprecation_message, deprecated_in, warn_from)
        return value


def deprecate(deprecated_in, warn_from=__version__, replaced_by=None, additional_info=''):
    """Deprecate a function, method or class."""

    def wrapper(deprecated_object, deprecation_message=None):
        if deprecation_message is None:
            deprecated_object_description = f'{deprecated_object.__module__}:{deprecated_object.__qualname__}'
            deprecation_message = _get_deprecation_message(deprecated_object_description, deprecated_in, replaced_by, additional_info)

        try:
            deprecated_object.__doc__ = _update_docstring(deprecation_message, deprecated_object.__doc__)
        except AttributeError:
            pass

        if hasattr(deprecated_object, 'mro'):
            deprecated_object.__init__ = wrapper(deprecated_object.__init__, deprecation_message)
            wrapped = deprecated_object

        else:
            @wraps(deprecated_object)
            def wrapped(*args, **kwargs):
                _warn(deprecation_message, deprecated_in, warn_from)
                return deprecated_object(*args, **kwargs)

        wrapped.deprecated_in = deprecated_in

        return wrapped

    return wrapper


def deprecate_module(module_locals, deprecated_in, warn_from=__version__, replaced_by=None, additional_info=''):
    """Deprecate a module."""
    deprecated_object_description = module_locals['__name__']
    deprecation_message = _get_deprecation_message(deprecated_object_description, deprecated_in, replaced_by, additional_info)
    module_locals['__doc__'] = _update_docstring(deprecation_message, module_locals['__doc__'])
    _warn(deprecation_message, deprecated_in, warn_from)


def _update_docstring(deprecation_message, docstring=None):
    if docstring is None:
        docstring = ''
    else:
        docstring += '\n\n'
    docstring += f"DEPRECATED - {deprecation_message}"
    return docstring


def _get_deprecation_message(deprecated_object_description, deprecated_in, replaced_by=None, additional_info=''):
    replacement = ''
    if replaced_by:
        replacement = f', and is replaced by {replaced_by}'
    if parse_version(__version__) < parse_version(deprecated_in):
        tense = ' will be'
        joiner = 'in'
    else:
        tense = ' is'
        joiner = 'since'
    deprecation_message = f'{deprecated_object_description}{tense} deprecated {joiner} version {deprecated_in}{replacement}{additional_info}'
    return deprecation_message


def _warn(deprecation_message, deprecated_in, warn_from):
    if BASE_VERSION >= parse_version(deprecated_in):
        raise DeprecatedError(deprecation_message)
    else:
        if BASE_VERSION >= parse_version(warn_from):
            warnings.warn(deprecation_message, APIDeprecationWarning)
        else:
            warnings.warn(deprecation_message, DeprecationWarning)


def is_deprecated(obj):
    """Check if an object is deprecated."""
    deprecated_in = getattr(obj, 'deprecated_in', None)
    if deprecated_in is None:
        # Check if it's a field
        if hasattr(obj, 'field_info'):
            deprecated_in = obj.field_info.extra.get('deprecated_in', None)
    return (deprecated_in is not None) and (BASE_VERSION >= parse_version(deprecated_in))



"""Add logger to a class."""

from fastapi_aad_auth.utilities import logging


class LoggingMixin:
    """Add logger to class based on name."""
    def __init__(self, *args, **kwargs):
        """Initialise the logger."""
        self._logger = None
        super().__init__(*args, **kwargs)

    @property
    def logger(self):
        """Get the logger object."""
        if self._logger is None:
            self._logger = logging.getLogger(self.__class__.__name__)
        return self._logger



"""Add not_authenticated error to a class."""
from fastapi import HTTPException
from starlette.status import HTTP_401_UNAUTHORIZED


class NotAuthenticatedMixin:
    """Provide an error for not authenticated error."""

    @property
    def not_authenticated(self):
        """Create an error for unauthenticated requests."""
        return HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED,
                    detail="Not authenticated",
                    headers={"WWW-Authenticate": "Bearer"},
                )



"""OAuth handlers."""
from fastapi_aad_auth.oauth.state import AuthenticationState  # noqa: F401
from fastapi_aad_auth.utilities import deprecate_module  # noqa: F401

_DEPRECATED_VERSION = '0.2.0'


deprecate_module(locals(), _DEPRECATED_VERSION)



"""Base OAuthBackend with token and session validators."""
from typing import List, Optional

from fastapi.security import OAuth2AuthorizationCodeBearer
from starlette.authentication import AuthCredentials, AuthenticationBackend, UnauthenticatedUser
from starlette.requests import Request

from fastapi_aad_auth._base.state import AuthenticationState
from fastapi_aad_auth._base.validators import SessionValidator, TokenValidator, Validator
from fastapi_aad_auth.mixins import LoggingMixin, NotAuthenticatedMixin
from fastapi_aad_auth.utilities import deprecate


class BaseOAuthBackend(NotAuthenticatedMixin, LoggingMixin, AuthenticationBackend):
    """Base OAuthBackend with token and session validators."""

    def __init__(self, validators: List[Validator], enabled: bool = True):
        """Initialise the validators."""
        super().__init__()
        self.enabled = enabled
        self.validators = validators[:]

    async def authenticate(self, request):
        """Authenticate a request.

        Required by starlette authentication middleware
        """
        state = self.check(request, allow_session=True)
        if state is None:
            return AuthCredentials([]), UnauthenticatedUser()
        return state.credentials, state.authenticated_user

    def is_authenticated(self, request: Request):
        """Check if a request is authenticated."""
        state = self.check(request, allow_session=True)
        return state is not None and state.is_authenticated()

    async def __call__(self, request: Request) -> Optional[AuthenticationState]:
        """Check/validate a request."""
        return self.check(request)

    def check(self, request: Request, allow_session=True) -> Optional[AuthenticationState]:
        """Check/validate a request."""
        state = None
        for validator in self.validators:
            if not allow_session and isinstance(validator, SessionValidator):
                self.logger.info('Skipping Session Validator as allow_session is False')
                continue
            state = validator.check(request)
            self.logger.debug(f'Authentication state {state} from validator {validator}')
            if state is not None and state.is_authenticated():
                break
        self.logger.info(f'Identified state {state}')
        return state

    def _iter_validators(self):
        """Iterate over authentication validators."""
        for validator in self.validators:
            yield validator

    def requires_auth(self, allow_session: bool = False):
        """Require authentication, use with fastapi Depends."""
        # This is a bit horrible, but is needed for fastapi to get this into OpenAPI (or similar) - it needs to be an OAuth2 object
        # We create this here "dynamically" for each endpoint, as we allow customisation on whether a session is permissible

        if self.enabled:
            class OAuthValidator(OAuth2AuthorizationCodeBearer):
                """OAuthValidator for API Auth."""

                def __init__(self_):
                    """Initialise the validator."""
                    token_validators = [u for u in self.validators if isinstance(u, TokenValidator)]
                    super().__init__(authorizationUrl=token_validators[0].model.flows.authorizationCode.authorizationUrl,
                                     tokenUrl=token_validators[0].model.flows.authorizationCode.tokenUrl,
                                     scopes=token_validators[0].model.flows.authorizationCode.scopes,
                                     refreshUrl=token_validators[0].model.flows.authorizationCode.refreshUrl)

                async def __call__(self_, request: Request):
                    """Validate a request."""
                    state = self.check(request, allow_session)
                    if state is None or not state.is_authenticated():
                        raise self.not_authenticated
                    return state

            return OAuthValidator()

        else:
            def noauth(request: Request):
                return AuthenticationState()

        return noauth

    @property  # type: ignore
    @deprecate('0.2.0', replaced_by=f'{__name__}:BaseOAuthBackend.requires_auth')
    def api_auth_scheme(self):
        """Get the API Authentication Schema."""
        return self.requires_auth()



"""Utilities."""
import importlib
from typing import List, Union

from pydantic.main import ModelMetaclass
from starlette.requests import Request

from fastapi_aad_auth.utilities import logging  # noqa: F401
from fastapi_aad_auth.utilities import urls  # noqa: F401
from fastapi_aad_auth.utilities.deprecate import DeprecatableFieldsMixin, deprecate, deprecate_module, DeprecatedField, is_deprecated  # noqa: F401


def is_interactive(request: Request):
    """Check if a request is from an interactive client."""
    return any([u in request.headers['user-agent'] for u in ['Mozilla', 'Gecko', 'Trident', 'WebKit', 'Presto', 'Edge', 'Blink']])


def bool_from_env(env_value: Union[bool, str]) -> bool:
    """Convert environment variable to boolean."""
    if isinstance(env_value, str):
        env_value = env_value.lower() in ['true', '1']
    return env_value


def list_from_env(env_value: Union[List[str], str]) -> List[str]:
    """Convert environment variable to list."""
    if isinstance(env_value, str):
        env_value = [u for u in env_value.split(',') if u]
    return env_value


def klass_from_str(value: str):
    """Convert an import path to a class."""
    if isinstance(value, str):
        if ':' in value:
            module_name, klass_name = value.split(':')
        else:
            split_path = value.split('.')
            module_name = '.'.join(split_path[:-1])
            klass_name = split_path[-1]
        module = importlib.import_module(module_name)
        value = getattr(module, klass_name)
    return value


def expand_doc(klass: ModelMetaclass) -> ModelMetaclass:
    """Expand pydantic model documentation to enable autodoc."""
    docs = ['', '', 'Keyword Args:']
    for name, field in klass.__fields__.items():  # type: ignore
        default_str = ''
        if field.default:
            default_str = f' [default: ``{field.default}``]'
        module = field.outer_type_.__module__
        if module != 'builtins':
            if hasattr(field.outer_type_, '__origin__'):
                type_ = f' ({field.outer_type_.__origin__.__name__}) '
            elif not hasattr(field.outer_type_, '__name__'):
                type_ = ''
            else:
                type_ = f' ({module}.{field.outer_type_.__name__}) '
        else:
            type_ = f' ({field.outer_type_.__name__}) '
        env_var = ''
        if 'env' in field.field_info.extra:
            env_var = f' (Can be set by ``{field.field_info.extra["env"]}`` environment variable)'
        docs.append(f'    {name}{type_}: {field.field_info.description}{default_str}{env_var}')
    if klass.__doc__ is None:
        klass.__doc__ = ''
    klass.__doc__ += '\n'.join(docs)
    return klass



"""Session based validator for interactive (UI) sessions."""
import fnmatch
from functools import partial
from typing import List, Optional

from itsdangerous import URLSafeSerializer

from fastapi_aad_auth._base.state import AuthenticationState
from fastapi_aad_auth._base.validators import Validator


REDIRECT_KEY = 'requested'


class SessionValidator(Validator):
    """Validator for session based authentication."""

    def __init__(self, session_serializer: URLSafeSerializer, ignore_redirect_routes: Optional[List[str]] = None, *args, **kwargs):
        """Initialise validator for session based authentication."""
        self._session_serializer = session_serializer
        self._ignore_redirect_routes = ignore_redirect_routes
        super().__init__(*args, **kwargs)  # type: ignore

    def get_state_from_session(self, request):
        """Get the session from the request."""
        auth_state = AuthenticationState.load_from_session(self._session_serializer, request.session)
        return auth_state

    def check(self, request):
        """Check the authentication from the request."""
        try:
            state = AuthenticationState.load_from_session(self._session_serializer, request.session)
        except Exception:
            state = AuthenticationState.as_unauthenticated(self._session_serializer, request.session)
            self.logger.exception('Error authenticating via session')
        return state

    def pop_post_auth_redirect(self, request):
        """Clear post-authentication redirects."""
        return request.session.pop(REDIRECT_KEY, request.query_params.get('redirect', '/'))

    def set_post_auth_redirect(self, request, redirect='/'):
        """Set post-authentication redirects."""
        if not self.is_valid_redirect(redirect):
            redirect = '/'

        request.session[REDIRECT_KEY] = redirect

    def is_valid_redirect(self, redirect):
        """Check if the redirect is not to endpoints that we don't want to redirect to."""
        if redirect is None:
            return False
        return not any(map(partial(fnmatch.fnmatch, redirect), self._ignore_redirect_routes))

    @staticmethod
    def get_session_serializer(secret, salt):
        """Get or Initialise the session serializer."""
        return URLSafeSerializer(secret, salt=salt)

    def logout(self, request):
        """Process a logout request."""
        AuthenticationState.logout(self._session_serializer, request.session)



"""Base Session Authenticator for interactive (UI) sessions."""
from starlette.requests import Request
from starlette.responses import RedirectResponse

from fastapi_aad_auth._base.state import AuthenticationState
from fastapi_aad_auth.errors import ConfigurationError
from fastapi_aad_auth.mixins import LoggingMixin


class SessionAuthenticator(LoggingMixin):
    """Authenticator for interactive (UI) sessions."""

    def __init__(self, session_validator, token_validator):
        """Initialise the session authenticator."""
        self._session_validator = session_validator
        self._token_validator = token_validator
        super().__init__()

    def redirect_if_authenticated(self, auth_state, redirect='/'):
        """Redirect to a target if authenticated."""
        if auth_state.is_authenticated():
            self.logger.info(f'Logged in, redirecting to {redirect}')
        else:
            redirect = '/login'
        return RedirectResponse(redirect)

    def redirect_to_provider_login(self, auth_state, request):
        """Redirect to the provider login."""
        self.logger.debug(f'state {auth_state}')
        auth_state.save_to_session(self._session_validator._session_serializer, request.session)
        authorization_url = self._get_authorization_url(request, auth_state.session_state)
        return RedirectResponse(authorization_url)

    def _get_authorization_url(self, request, session_state):
        raise NotImplementedError('Implement in specific subclass')

    def process_login_request(self, request, force=False, redirect='/'):
        """Process the provider login request."""
        self.logger.debug(f'Logging in - request url {request.url}')
        auth_state = self._session_validator.get_state_from_session(request)
        if auth_state.is_authenticated() and not force:
            self.logger.debug(f'Authenticated - redirecting {auth_state}')
            response = self.redirect_if_authenticated(auth_state)
        else:
            # Set the redirect parameter here
            self._session_validator.set_post_auth_redirect(request, request.query_params.get('redirect', redirect))
            self.logger.debug(f'No Auth state - redirecting to provider login {auth_state}')
            response = self.redirect_to_provider_login(auth_state, request)
        return response

    def process_login_callback(self, request):
        """Process the provider login callback."""
        if 'error' in request.query_params:
            error_args = [request.query_params['error'], ]
            if 'error_description' in request.query_params:
                error_args.append(request.query_params['error_description'])
            raise ConfigurationError(*error_args)
        code = request.query_params.get('code', None)
        state = request.query_params.get('state', None)
        if state is None or code is None:
            return  # not authenticated
        auth_state = self._session_validator.get_state_from_session(request)
        auth_state.check_session_state(state)
        token = self._process_code(request, auth_state, code)
        user = self._get_user_from_token(token)
        authenticated_state = AuthenticationState.authenticate_as(user, self._session_validator._session_serializer, request.session)
        redirect = self._session_validator.pop_post_auth_redirect(request)
        return self.redirect_if_authenticated(authenticated_state, redirect=redirect)

    def _process_code(self, request, auth_state, code):
        raise NotImplementedError('Implement in subclass')

    def get_access_token(self, user, scopes=None, app_scopes=True):
        """Get the access token for the user."""
        raise NotImplementedError('Implement in subclass')

    def get_access_token_from_request(self, request: Request):
        """Get the access token from a request object."""
        auth_state = self._session_validator.get_state_from_session(request)
        if auth_state.is_authenticated():
            return self.get_access_token(auth_state.user)['access_token']
        return None

    def get_user_from_request(self, request: Request):
        """Get the user from a request object."""
        auth_state = self._session_validator.get_state_from_session(request)
        return auth_state.user

    def _get_user_from_token(self, token, options=None):
        validated_claims = self._token_validator.validate_token(token, options=options)
        return self._token_validator._get_user_from_claims(validated_claims)

    def get_login_button(self, url, post_redirect='/'):
        """Get a UI login button."""
        url = self._add_redirect_to_url(url, post_redirect)
        return f'<a class="btn btn-lg btn-primary btn-block col-8 offset-md-2" href="{url}">Sign in</a>'

    def logout(self, request):
        """Process a logout request if any special behaviour required."""
        pass

    def pop_post_auth_redirect(self, *args, **kwargs):
        """Clear post-authentication redirects."""
        return self._session_validator.pop_post_auth_redirect(*args, **kwargs)

    def set_post_auth_redirect(self, *args, **kwargs):
        """Set post-authentication redirects."""
        self._session_validator.set_post_auth_redirect(*args, **kwargs)



"""AAD handlers."""
from fastapi_aad_auth._base.validators import SessionValidator as _SessionValidator
from fastapi_aad_auth.auth import Authenticator as _Authenticator
from fastapi_aad_auth.config import AADConfig, Config  # noqa: F401
from fastapi_aad_auth.providers.aad import AADProvider as _AADProvider
from fastapi_aad_auth.utilities import deprecate, deprecate_module

_DEPRECATED_VERSION = '0.2.0'


deprecate_module(locals(), _DEPRECATED_VERSION, replaced_by=f'{_AADProvider.__module__} and {_Authenticator.__module__}')


@deprecate(_DEPRECATED_VERSION, replaced_by=f'{_AADProvider.__module__}:{_AADProvider.__name__}')
class AADOAuthBackend(_AADProvider):  # noqa: D101
    __doc__ = _AADProvider.__doc__

    @classmethod
    @deprecate(_DEPRECATED_VERSION, replaced_by=f'{_AADProvider.__module__}:{_AADProvider.__name__}.from_config and {_Authenticator.__module__}:{_Authenticator.__name__}')
    def from_config(cls, config: Config):
        """Load the auth backend from a config.

        Args:
            config: Loaded configuration

        Keyword Args:
            user_klass: The class to use as a user
        """
        session_validator = _SessionValidator.get_session_serializer(config.auth_session.secret.get_secret_value(),
                                                                     config.auth_session.salt.get_secret_value())
        provider = super().from_config(session_validator, config, config.aad)
        return provider.auth_backend



from abc import abstractmethod

from starlette.requests import Request

from fastapi_aad_auth._base.state import AuthenticationState
from fastapi_aad_auth.mixins import LoggingMixin, NotAuthenticatedMixin


class Validator(NotAuthenticatedMixin, LoggingMixin):
    """Base Validator Class."""

    @abstractmethod
    def check(self, request: Request) -> AuthenticationState:
        """Check a request."""
        raise NotImplementedError('Implement in subclass')

    async def __call__(self, request: Request) -> AuthenticationState:  # type: ignore
        """Validate the request authentication.

        Returns an AuthenticationState object or raises an Unauthorized error
        """
        result = self.check(request)
        self.logger.info(f'Identified state {result}')
        if not result.is_authenticated():
            raise self.not_authenticated
        return result



from typing import List, Optional

from starlette.requests import Request
from starlette.responses import RedirectResponse
from starlette.routing import Route

from fastapi_aad_auth._base.authenticators import SessionAuthenticator
from fastapi_aad_auth._base.validators import Validator
from fastapi_aad_auth.mixins import LoggingMixin
from fastapi_aad_auth.utilities import urls


class Provider(LoggingMixin):
    """Authentication Provider."""
    name: Optional[str] = None

    def __init__(self, validators: List[Validator], authenticator: SessionAuthenticator, enabled: bool = True, oauth_base_route: str = '/oauth'):
        """Initialise the authentication provider."""
        self.validators = validators
        self.authenticator = authenticator
        self.enabled = enabled
        self.oauth_base_route = oauth_base_route
        self._login_url = None
        self._redirect_url = None
        super().__init__()

    def get_routes(self, noauth_redirect='/'):
        """Get the authenticator routes."""

        async def login(request: Request):
            self.logger.debug(f'Logging in with {self.name} - request url {request.url}')
            if self.enabled:
                self.logger.debug(f'Auth {request.auth}')
                return self.authenticator.process_login_request(request)
            else:
                self.logger.debug('Auth not enabled')
                return RedirectResponse(noauth_redirect)

        async def login_callback(request: Request):
            self.logger.info(f'Processing login callback for {self.name}')
            self.logger.debug(f'request url {request.url}')
            if self.enabled:
                return self.authenticator.process_login_callback(request)
            else:
                self.logger.debug('Auth not enabled')
                return RedirectResponse(noauth_redirect)

        routes = [Route(self.login_url,
                        endpoint=login, methods=['GET'], name=f'oauth_login_{self.name}'),
                  Route(self.redirect_url,
                        endpoint=login_callback, methods=['GET'], name=f'oauth_login_{self.name}_callback')]
        return routes

    def _build_oauth_url(self, oauth_base_route, route):
        return urls.append(oauth_base_route, self.name, route)

    def logout(self, request):
        """Logout from the authenticator."""
        pass

    @property
    def login_url(self):
        """Get the login url."""
        if self._login_url is None:
            self._login_url = self._build_oauth_url(self.oauth_base_route, 'login')
        return self._login_url

    @property
    def redirect_url(self):
        """Get the login redirect url."""
        if self._redirect_url is None:
            self._redirect_url = self._build_oauth_url(self.oauth_base_route, 'redirect')
        return self._redirect_url



"""AAD OAuth handlers."""

import base64
from enum import Enum
from typing import List, Optional, Union

from authlib.jose import errors as jwt_errors, jwk, jwt
from authlib.jose.util import extract_header
from cryptography.hazmat.primitives import serialization
import msal
from pkg_resources import resource_string
from pydantic import BaseSettings as _BaseSettings, Field, HttpUrl, PrivateAttr, SecretStr, validator
import requests
from starlette.middleware.authentication import AuthenticationError
from starlette.requests import Request

from fastapi_aad_auth._base.authenticators import SessionAuthenticator
from fastapi_aad_auth._base.provider import Provider
from fastapi_aad_auth._base.state import User
from fastapi_aad_auth._base.validators import SessionValidator, TokenValidator
from fastapi_aad_auth.errors import ConfigurationError
from fastapi_aad_auth.utilities import bool_from_env,  DeprecatableFieldsMixin, expand_doc, is_deprecated, list_from_env, urls


class TokenType(Enum):
    """Type of token to use."""
    access = 'access_token'
    id = 'id_token'


class BaseSettings(DeprecatableFieldsMixin, _BaseSettings):
    """Base Settings with Deprecatable Fields."""
    pass


class AADSessionAuthenticator(SessionAuthenticator):
    """AAD Authenticator for interactive (UI) sessions."""

    def __init__(
            self,
            session_validator,
            token_validator,
            client_id,
            tenant_id,
            redirect_path='/oauth/aad/redirect',
            prompt=None,
            client_secret=None,
            scopes=None,
            redirect_uri=None,
            domain_hint=None,
            token_type=TokenType.access):
        """Initialise AAD Authenticator for interactive (UI) sessions."""
        super().__init__(session_validator, token_validator)
        self._redirect_path = redirect_path
        self._redirect_uri = redirect_uri
        self._domain_hint = domain_hint
        self._prompt = prompt
        self.client_id = client_id
        if scopes is None:
            scopes = [f'{self.client_id}/openid']
        self.logger.info(f'Scopes {scopes}')
        self._scopes = scopes
        if isinstance(token_type, Enum):
            token_type = token_type.value
        self._token_type = token_type
        self._authority = f'https://login.microsoftonline.com/{tenant_id}'

        if client_secret is not None:
            self.logger.info('Client secret provided, using Confidential Client')
            self.msal_application = msal.ConfidentialClientApplication(
                client_id,
                authority=self._authority,
                client_credential=client_secret)
        else:
            self.logger.info('Client secret not provided, using Public Client')
            self.msal_application = msal.PublicClientApplication(
                client_id,
                authority=self._authority)

    def _build_redirect_uri(self, request: Request):
        if self._redirect_uri:
            redirect_uri = self._redirect_uri
        else:
            if request.url.port is None or (request.url.port == 80 and request.url.scheme == 'http') or (request.url.port == 443 and request.url.scheme == 'https'):
                port = ''
            else:
                port = f':{request.url.port}'
            redirect_uri = f'{request.url.scheme}://{request.url.hostname}{port}{self._redirect_path}'
        return redirect_uri

    def _process_code(self, request: Request, auth_state, code):
        # Let's build up the redirect_uri
        result = self.msal_application.acquire_token_by_authorization_code(code, scopes=self._scopes,
                                                                           redirect_uri=self._build_redirect_uri(request))
        self.logger.debug(f'Result {result}')
        if 'error' in result and result['error']:
            error_args = [result['error']]
            if 'error_description' in result:
                error_args.append(result['error_description'])
            raise ConfigurationError(*error_args)
        return result[self._token_type]

    def _get_user_from_token(self, token, options=None):
        if options is None:
            options = self._token_validator._claims_options
        options.pop('azp', None)
        options.pop('appid', None)
        return super()._get_user_from_token(token, options=options)

    def _get_authorization_url(self, request, session_state):
        return self.msal_application.get_authorization_request_url(self._scopes,
                                                                   state=session_state,
                                                                   claims_challenge='{"id_token": {"roles": {"essential": true} } }',
                                                                   redirect_uri=self._build_redirect_uri(request),
                                                                   prompt=self._prompt,
                                                                   domain_hint=self._domain_hint)

    def get_access_token(self, user, scopes=None, app_scopes=True):
        """Get the access token for the user."""
        result = None
        account = None
        if scopes is None:
            scopes = self._scopes
        elif app_scopes:
            scopes = self.as_app_scopes(scopes)
        if user.username:
            account = self.msal_application.get_accounts(user.username)
        if account:
            account = account[0]
            self.logger.info(account)
            # This needs you to register the scopes in the app registration
            result = self.msal_application.acquire_token_silent_with_error(scopes=scopes, account=account)
            self.logger.info(f'Acquired Token: {result}')
        if result is None:
            raise ValueError('Token not found')
        elif 'error' in result:
            raise ConfigurationError(result['error'], result['error_description'])
        else:
            return {'token_type': result['token_type'],
                    'expires_in': result['expires_in'],
                    'access_token': result['access_token']}

    def as_app_scopes(self, scopes):
        """Add the application client id to the scopes so that the tokens are valid for this app."""
        if self.client_id not in scopes[0]:
            scopes[0] = f'{self.client_id}/{scopes[0]}'
        return scopes


class AADTokenValidator(TokenValidator):
    """Validator for AAD token based authentication."""

    def __init__(self,
                 client_id: str,
                 tenant_id: str,
                 api_audience: str = None,
                 scheme_name: str = None,
                 scopes: dict = None,
                 auto_error: bool = False,
                 enabled: bool = True,
                 use_pkce: bool = True,
                 strict: bool = True,
                 client_app_ids: Optional[List[str]] = None,
                 user_klass: type = User):
        """Initialise validator for AAD token based authentication."""
        authorization_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/authorize"
        token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        self.key_url = f"https://login.microsoftonline.com/{tenant_id}/discovery/v2.0/keys"
        self.tenant_id = tenant_id
        super().__init__(client_id=client_id, authorizationUrl=authorization_url, tokenUrl=token_url, api_audience=api_audience, scheme_name=scheme_name,
                         scopes=scopes, auto_error=auto_error, enabled=enabled, use_pkce=use_pkce, user_klass=user_klass)
        self.strict = strict
        if client_app_ids is None:
            client_app_ids = []
        self.client_app_ids = client_app_ids

    def _get_ms_jwk(self, token):
        try:
            self.logger.info(f'Getting signing keys from {self.key_url}')
            jwks = requests.get(self.key_url).json()
            token_header = token.split(".")[0].encode()
            unverified_header = extract_header(token_header, jwt_errors.DecodeError)
            for key in jwks["keys"]:
                if key["kid"] == unverified_header["kid"]:
                    self.logger.info(f'Identified key {key["kid"]}')
                    return jwk.loads(key)
        except jwt_errors.DecodeError:
            self.logger.exception('Error parsing signing keys')
        raise AuthenticationError("Unable to parse signing keys")

    def _decode_token(self, token):
        jwk_ = self._get_ms_jwk(token)
        claims = None
        self.logger.debug(f'Key is {jwk_}')
        try:
            if hasattr(jwk, 'public_bytes'):
                key = jwk_.public_bytes(serialization.Encoding.PEM, serialization.PublicFormat.PKCS1)
            else:
                key = jwk_.raw_key.public_bytes(serialization.Encoding.PEM, serialization.PublicFormat.PKCS1)
            self.logger.debug(f'Processed Key: {key}')
            claims = jwt.decode(
                token,
                key,
            )
        except Exception:
            self.logger.exception('Unable to parse error')
            raise AuthenticationError("Unable to parse authentication token")
        return claims

    def _validate_claims(self, claims, options=None):
        if options is None:
            options = self._claims_options
        # We need to do some 1.0/2.0 handling because it doesn't seem to work properly
        # TODO: validate whether we want this claim here?
        if 'appid' in options and 'azp' in options:
            if 'appid' not in claims:
                options.pop('appid')
            elif 'azp'not in claims:
                options.pop('azp')
            if not ('appid' in claims or 'azp' in claims):
                if self.strict:
                    self.logger.error('No appid/azp claims found in token')
                    raise AuthenticationError('No appid/azp claims found in token')
                else:
                    self.logger.warning('No appid/azp claims found in token - we are ignoring for now')
        return super()._validate_claims(claims, options)

    @property
    def _claims_options(self):
        options = super()._claims_options
        options["iss"] = {"essential": True, "values": [f"https://sts.windows.net/{self.tenant_id}/", f"https://login.microsoftonline.com/{self.tenant_id}/v2.0"]}
        options["aud"] = {"essential": True, "values": [self.api_audience] + [self.client_id] + self.client_app_ids}
        options["azp"] = {"essential": True, "values": [self.client_id] + self.client_app_ids}
        options["appid"] = {"essential": True, "values": [self.client_id] + self.client_app_ids}
        self.logger.debug(f'Claims options {options}')
        return options

    def _get_user_from_claims(self, claims):
        self.logger.debug(f'Processing claims: {claims}')
        username_key = 'preferred_username'
        if username_key not in claims:
            username_key = 'unique_name'
        if 'name' not in claims and 'appid' in claims:
            # This is an application/service principal
            user = self._user_klass(name=claims['appid'], email='', username=claims['appid'], groups=claims.get('groups', None), roles=claims.get('roles', None), scopes=claims.get('scp', None))
        else:
            user = self._user_klass(name=claims['name'], email=claims[username_key], username=claims[username_key], groups=claims.get('groups', None), roles=claims.get('roles', None), scopes=claims.get('scp', None))
        return user


class AADProvider(Provider):
    """fastapi auth backend for Azure Active Directory."""
    name: str = 'aad'

    def __init__(
            self,
            session_validator: SessionValidator,
            client_id: str,
            tenant_id: str,
            prompt: Optional[str] = None,
            client_secret: Optional[str] = None,
            scopes: Optional[List[str]] = None,
            enabled: bool = True,
            client_app_ids: Optional[List[str]] = None,
            strict_token: bool = True,
            api_audience: Optional[str] = None,
            redirect_uri: Optional[str] = None,
            domain_hint: Optional[str] = None,
            user_klass: type = User,
            oauth_base_route: str = '/oauth',
            token_type: Union[str, TokenType] = TokenType.access):
        """Initialise the auth backend.

        Args:
            session_serializer: Session serializer object
            client_id: Client ID from Azure App Registration
            tenant_id: Tenant ID to connect to for Azure App Registration

        Keyword Args:
            prompt: Prompt options for Azure AD
            client_secret: Client secret value
            scopes: Additional scopes requested
            enabled: Boolean flag to enable this backend
            client_app_ids: List of client apps to accept tokens from
            strict_token: Strictly evaluate token
            api_audience: Api Audience declared in Azure AD App registration
            redirect_uri: Full URI for post authentication callbacks
            domain_hint: Hint for the domain
            user_klass: Class to use as a user.
        """
        redirect_path = self._build_oauth_url(oauth_base_route, 'redirect')
        token_validator = AADTokenValidator(client_id=client_id, tenant_id=tenant_id, api_audience=api_audience,
                                            client_app_ids=client_app_ids, scopes={}, enabled=enabled, strict=strict_token,
                                            user_klass=user_klass)
        session_authenticator = AADSessionAuthenticator(session_validator=session_validator, token_validator=token_validator,
                                                        client_id=client_id, tenant_id=tenant_id, redirect_path=redirect_path,
                                                        prompt=prompt, client_secret=client_secret, scopes=scopes,
                                                        redirect_uri=redirect_uri, domain_hint=domain_hint, token_type=token_type)
        super().__init__(validators=[token_validator], authenticator=session_authenticator, enabled=enabled, oauth_base_route=oauth_base_route)

    @classmethod
    def from_config(cls, session_validator, config, provider_config, user_klass: Optional[type] = None):
        """Load the auth backend from a config.

        Args:
            session_validator (SessionValidator): the session validator to use
            config: Loaded configuration

        Keyword Args:
            user_klass: The class to use as a user
        """
        client_secret = provider_config.client_secret
        if client_secret is not None:
            client_secret = client_secret.get_secret_value()  # type: ignore

        if user_klass is None:
            user_klass = config.user_klass

        obj = cls(session_validator=session_validator, client_id=provider_config.client_id.get_secret_value(),
                  tenant_id=provider_config.tenant_id.get_secret_value(),
                  client_secret=client_secret, enabled=config.enabled,   # type: ignore
                  scopes=provider_config.scopes, client_app_ids=provider_config.client_app_ids,
                  strict_token=provider_config.strict, api_audience=provider_config.api_audience,
                  prompt=provider_config.prompt, domain_hint=provider_config.domain_hint,
                  redirect_uri=provider_config.redirect_uri, user_klass=user_klass,
                  oauth_base_route=config.routing.oauth_base_route, token_type=provider_config.token_type)
        # We need to override the login and redirect etc until it is deprecated
        if hasattr(config.routing, 'login_path') and config.routing.login_path and not is_deprecated(config.routing.__fields__['login_path']):
            obj._login_url = config.routing.login_path
        if hasattr(config.routing, 'login_redirect_path') and config.routing.login_redirect_path and not is_deprecated(config.routing.__fields__['login_redirect_path']):
            obj._redirect_url = config.routing.login_redirect_path
        obj.authenticator._redirect_path = obj.redirect_url  # type: ignore
        return obj

    def get_login_button(self, post_redirect='/'):
        """Get the AAD Login Button."""
        url = urls.with_redirect(self.login_url, post_redirect)
        logo = base64.b64encode(resource_string('fastapi_aad_auth.providers', 'ms-logo.png')).decode()
        return f'<a class="btn btn-lg btn-light btn-ms" href="{url}"><div class="row align-items-center justify-center login-ms"><img alt="Microsoft Logo" class="rounded splash-ms" src="data:image/png;base64,{logo}" />Sign in with Microsoft Work Account</div></a>'


@expand_doc
class AADConfig(BaseSettings):
    """Configuration for the AAD application.

    Includes expected claims, application registration, etc.

    Can also provide additional client application ids to accept.

    A list of roles can be provided to accept (requires configuring the
    roles in the AAD application registration manifest)
    """
    client_id: SecretStr = Field(..., description="Application Registration Client ID", env='AAD_CLIENT_ID')
    tenant_id: SecretStr = Field(..., description="Application Registration Tenant ID", env='AAD_TENANT_ID')
    client_secret: Optional[SecretStr] = Field(None, description="Application Registration Client Secret (if required)", env='AAD_CLIENT_SECRET')
    scopes: Optional[List[str]] = Field(None, description="Additional scopes requested - if the scope is not configured to the application this will throw an error when validating the token")
    client_app_ids: Optional[List[str]] = Field(None, description="Additional Client App IDs to accept tokens from (when running as a backend service)",
                                                env='AAD_CLIENT_APP_IDS')
    strict: bool = Field(True, description="Check that all claims are provided", env='AAD_STRICT_CLAIM_CHECK')
    api_audience: Optional[str] = Field(None, description="Corresponds to the Application ID URI - used for token validation, defaults to api://{client_id}",
                                        env='AAD_API_AUDIENCE')
    redirect_uri: Optional[HttpUrl] = Field(None, description="The redirect URI to use - overwrites the default path handling etc",
                                            env='AAD_REDIRECT_URI')
    prompt: Optional[str] = Field(None, description="AAD prompt to request", env='AAD_PROMPT')
    domain_hint: Optional[str] = Field(None, description="AAD domain hint", env='AAD_DOMAIN_HINT')
    roles: Optional[List[str]] = Field(None, description="AAD roles required in claims", env='AAD_ROLES')
    token_type: TokenType = Field(TokenType.access, description='The AAD token type to use to validate (we should use the access token if it is configured, unless we are acting as a pure UI component')
    _provider_klass: type = PrivateAttr(AADProvider)

    class Config:  # noqa D106
        env_file = '.env'

    _validate_strict = validator('strict', allow_reuse=True)(bool_from_env)
    _validate_client_app_ids = validator('client_app_ids', allow_reuse=True)(list_from_env)
    _validate_roles = validator('roles', allow_reuse=True)(list_from_env)



"""fastapi_aad_auth errors."""
from pathlib import Path
from typing import Dict, Optional

from starlette.authentication import AuthenticationError
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse, Response
from starlette.templating import Jinja2Templates

from fastapi_aad_auth.utilities import is_interactive, urls
from fastapi_aad_auth.utilities.logging import getLogger

logger = getLogger(__name__)


def base_error_handler(request: Request, exception: Exception, error_type: str, error_message: str, templates: Jinja2Templates, template_path: Path, context: Optional[Dict] = None, status_code: int = 500) -> Response:
    """Handle Error as JSON or HTML response depending on request type."""
    logger.warning(f'Handling error {exception}')
    if is_interactive(request):
        response = ui_error_handler(request, exception, error_type, error_message, templates, template_path, context, status_code)
    else:
        response = json_error_handler(error_message, status_code)
    logger.debug(f'Response {response}')
    return response


def json_error_handler(error_message: str, status_code: int = 500) -> JSONResponse:
    """Handle error as a JSON."""
    logger.info('Non-Interactive environment so returning JSON message')

    return JSONResponse(   # type: ignore
        status_code=status_code,
        content={"message": error_message}
    )


def redirect_error_handler(redirect_path: str, exception: Exception, **kwargs) -> RedirectResponse:
    """Handle error as a redirect with error info in the query parameters."""
    return RedirectResponse(urls.with_query_params(redirect_path, error=exception, **kwargs))


def ui_error_handler(request: Request, exception: Exception, error_type: str, error_message: str, templates: Jinja2Templates, template_path: Path, context: Optional[Dict] = None, status_code: int = 500) -> Response:
    """Return a UI view of the error."""
    logger.info('Interactive environment so returning error template')
    logger.debug(f'Path: {template_path}')
    logger.debug(f'Exception: {exception}')
    if context is None:
        context = {}
    error_context = context.copy()
    error = exception
    detail = ''
    if exception.args:
        logger.info('Getting args')
        error = exception.args[0]
    if len(exception.args) > 1:
        detail = exception.args[1]
    error_context.update({'error': str(error),
                          'status_code': str(status_code),
                          'error_type': error_type,
                          'error_description': error_message,
                          'error_detail': str(detail),
                          'request': request})  # type: ignore
    logger.debug(f'Error context: {error_context}')
    return templates.TemplateResponse(template_path.name,
                                      error_context,
                                      status_code=status_code)


class ConfigurationError(Exception):
    """Misconfigured application."""


class AuthorisationError(AuthenticationError):
    """Not Authorised to access this resource."""




from fastapi_aad_auth._base.validators.base import Validator  # noqa: F401
from fastapi_aad_auth._base.validators.session import SessionValidator  # noqa: F401
from fastapi_aad_auth._base.validators.token import TokenValidator  # noqa: F401



from fastapi_aad_auth.auth import Authenticator  # noqa F401
from fastapi_aad_auth.config import Config  # noqa F401
from fastapi_aad_auth._base.state import AuthenticationState  # noqa F401
from fastapi_aad_auth._version import get_versions

__version__ = get_versions()['version']
del get_versions



"""URL utilities."""
import logging

from starlette.datastructures import URL


logger = logging.getLogger(__name__)


def with_redirect(url, post_redirect=None):
    """Append a redirect query parameter."""
    if post_redirect is not None:
        url = with_query_params(url, redirect=post_redirect)
    return url


def append(base_url, *args):
    """Append paths together."""
    extension = '/'.join([u.strip('/') for u in args])
    if extension:
        url = base_url.rstrip('/')+'/'+extension
    else:
        url = base_url
    return url


def with_query_params(url, **query_params):
    """Add query parameters to a url."""
    logger.debug(f'Adding {query_params} to {url}')
    parsed_url = URL(url)
    logger.debug(f'Existing query params {parsed_url.query}')
    new_url = parsed_url.include_query_params(**query_params)
    logger.debug(f'Updated query params {new_url.query}')
    logger.debug(f'Updated url {new_url}')
    return str(new_url)
